#! /bin/sh

I="<6>"
W="<4>"
E="<3>"

warn() { echo "${W}warning: $*" >&2; }
info() { echo "${I}$*" >&2; }

echo
info "Processing INITRAMFS"

if [ -f /config0 ]; then . /config0; fi

chk() {
  while ! "$@"; do
    warn "the following command failed:"
    warn "$*"
    p="** Continue(Ignore)/Shell/Retry (C/s/r)? "
    while : ; do
      if ! read -t 10 -p "$p" x 2>&1; then
        echo "(timeout, continuing)"
        return 1
      fi
      case "$x" in
        [Ss!]*) /bin/sh 2>&1 ;;
        [Rr]*) break;;
        [CcIi]*|"") return 1;;
        *) echo "(unrecognized response)";;
      esac
    done
  done
}

modload() {
  [ ! -d /sys/module/$1 ] || return 0
  info "loading module $1"
  chk modprobe $1
}

runshell() {
 case "$shell" in
  [Yy1]* ) shell=/bin/sh ;;
  [Nn0]* ) shell= ;;
 esac
 if [ -n "$shell" ]; then
   echo "Running shell $*.  Exit from shell to continue"
   $shell 2>&1
 fi
}

findblkdev() {
  dev=$1 what=$2
  major= minor=
  case "$dev" in
    /dev/*) ;;
    [Uu][Uu][Ii][Dd]=*) dev=/dev/disk/by-uuid/${dev#*=} ;;
    [Ll][Aa][Be][Ee][Ll]=*) dev=/dev/disk/by-label/${dev#*=} ;;
    *:*) major=${dev%:*}; minor=${dev#*:} ;;
    *) dev=/dev/$dev ;;
  esac
}

mkdir -p /dev /proc /sys /run /root
[ -d /proc/self/ ] ||
  chk mount -n -t proc proc /proc
[ -d /sys/bus/ ] ||
  chk mount -n -t sysfs sysfs /sys
chk mount -n -t tmpfs -o nosuid,size=20%,mode=0755 runfs /run
[ -c /dev/kmsg ] ||
  mount -t devtmpfs -o nosuid,noexec,size=1M udev /dev

# Parse command line
read x < /proc/cmdline

for key in $x $*; do
  case "$key" in
    *=*) val="${key#*=}" key="${key%%=*}" ;;
    shell) shell=y; continue;;
    bootdebug) set -x ;;
    ro|rw)
      rootflags=$(echo ,$rootflags, | sed -e 's|,r[wo],|,|' -e 's|^,||' -e 's|,$||')
      if [ ! "$rootflags" ]; then rootflags=$key
      else rootflags="$rootflags,$key"
      fi
      continue;;
    single|S|[1-5]|emergency|auto) initargs=$key; continue;;
    noresume) resume=; continue;;
    rootwait) rootwait=10; continue;;
    -b) initargs=emergency; continue;;
    -s) initargs=single; continue;;
    -a) initargs=auto; continue;;
    *) continue;;
  esac
  case "$key" in
  rootfstype|rootflags|init|ip|shell|bootrc|hostname|resume|ovl|rootsquash)
    eval $key=\$val ;;
  rootpath) root=$val ;;
  rootargs) rootflags=$val ;;
  rootfs) rootfstype=$val ;;
  root)
    case "$val" in
      ram|ram0|/dev/ram|/dev/ram0|100)
        # ignore root=/dev/ram0
        warn "ignoring root=$val"
        continue;;
      *) root=$val ;;
    esac
    ;;
  rootwait|rootdelay) eval $key=\$val ;;
  #nfsaddr ) ip=$val ;;
  load|noload)
    if [ -z "$val" -o ".$val" = .none ]; then eval $key=
    else eval $key=\"\$$key\ \$val\"
    fi
    ;;
  md[0-9]|md[1-0][0-9]|md[12][0-9][0-9])
    case "$val" in
      +*) eval "$key=\"\$$key,\${val#+}\"" ;;
      *)  eval "$key=\$val" ;;
    esac
    md="$md $key"
    ;;
  md?*) warn "unrecognized parameter \"$key=$val\"" ;;
  md)
    case "$val" in
      "") md= ;; # reset
      *) warn "unrecognized parameter \"$key=$val\"" ;;
    esac
    ;;
  BOOTIF|bootif) # added by pxelinux
    bootif=$val
    ;;
  bridge|network.bridge)
    bridge=$val
    ;;
  esac
done

# setup blacklist entries from noload stuff
for mod in $noload; do
  echo blacklist $mod
done >> /etc/modprobe.conf

# Load modules if any
for key in $load; do
  case "$noload " in
    *" $key "*) ;;
    *) modload "$key" ;;
  esac
done

# Load device drivers.
[ ! -w /sys/kernel/uevent_helper ] ||
  echo > /sys/kernel/uevent_helper
chk udevd --daemon --resolve-names=never
udevadm trigger --action=add
udevadm settle

[ -z "$rootdelay" ] ||
  sleep $rootdelay

if [ -n "$resume" ]; then
  if [ -x /bin/resume ]; then
    runshell "before resume (from $dev)"
    /bin/resume "$resume"
  elif [ -f /sys/power/resume ]; then
    findblkdev $resume resume
    if [ -b $dev ] && set -- $(stat -c "%t %T" $dev); then
      devno=$((0x$1)):$((0x$2))	# convert from hex to dec
      runshell "before resume (from $devno)"
      info "trying to resume from $dev ($devno)"
      echo -n $devno > /sys/power/resume
    fi
    warn "resume from $dev failed, continuing"
  else
    warn "kernel does not support resuming, continuing"
  fi
fi

# Configure NIC
if [ -n "$ip" ]; then
  ip="$ip:"
  client_ip=${ip%%:*} ip=${ip#*:}
  server_ip=${ip%%:*} ip=${ip#*:}
  gw_ip=${ip%%:*} ip=${ip#*:}
  netmask=${ip%%:*} ip=${ip#*:}
  [ -n "$hostname" ] || hostname=${ip%%:*} ip=${ip#*:}
  iface=eth0
  if [ -n "$bootif" ]; then
    bootif="$(echo "$bootif" | sed -ey/-/:/ -e's/^0\?1://')"
    case "$bootif" in
      *:*:*:*:*:*:*)
        warn "unsupported net ${bootif%%:*}, using inet"
        bootif=${bootif#*:}
        ;;
    esac
    set -- $(ifconfig -a | grep -i "$bootif")
    if [ -n "$1" ]; then iface=$1
    else warn "unable to find netif for $bootif, using $iface"
    fi
  fi
  info "configuring network interface $iface: $client_ip/$netmask"
  chk ifconfig $iface inet "$client_ip" netmask "$netmask" up
  case "$gw_ip" in
    "" | 0*) ;;
    *) chk route add default gw $gw_ip ;;
  esac
fi
if [ -n "$hostname" ]; then
  info "setting hostname to \"$hostname\""
  echo "$hostname" > /proc/sys/kernel/hostname
fi

# Start RAID arrays if any
mdl=
for dev in $md; do
  eval c=\$$dev
  if [ -n "$c" ]; then
    case " $mdl " in
     *\ $dev\ *) ;;
     *) mdl="$mdl $dev" ;;
    esac
  fi
done
md="${mdl# }"
if [ -n "$md" ]; then
  [ -f /proc/mdstat ] || modprobe md-mod
  for dev in $md; do
    eval c=\$$dev,
    uuid= level= opt= devices=
    while [ -n "$c" ]; do
      x=${c%%,*}; c=${c#*,}
      case "$x" in
        "") continue ;;
        uuid=*) uuid=--$x ;;
        level=[01456]|level=10|level=linear)
          level=--$x;;
        *:*:*) uuid=--uuid=$x ;;
        -*) opt="$opt $x" ;;
        /dev/*) devices="$devices $x" ;;
        *) warn "unrecognized $dev parameter \"$x\""; continue;;
      esac
    done
    info "configuring raid array $dev"
    chk mdadm --assemble /dev/$dev $level $uuid $devices $opt
  done
fi

if [ "x$root" = "xnone" ]; then
  :
elif [ "x${root#nbd:}" != "x${root}" ]; then
  # root=nbd:server:port
  # only old-style for now, since busybox does not support new style

  modprobe nbd nbds_max=2
  root=${root#nbd:}
  chk nbd-client ${root//:/ } /dev/nbd0
  root=/dev/nbd0

else

 # Determine root device
  findblkdev "$root" root
  root=$dev
  case "$rootfstype" in
    isofs|iso|cdrom) rootfstype=iso9660 ;;
  esac

fi

case "$root" in
  /dev/*)
    if [ ! -e "$root" ]; then
      if [ -n "$rootwait" ]; then
        info "waiting for $root to appear (${rootwait}s max)"
        n=0 rootwait=${rootwait}0
        while [ $n -lt $rootwait -a ! -e $root ]; do
          sleep 0.1; n=$(($n+1))
        done
        [ -e "$root" ] || warn "root device $root not appeared in $max sec"
      else
        warn "root device $root does not exist"
      fi
    fi
    if [ -L "$root" -a -e "$root" ]; then
      root=$(readlink -f "$root")
    fi
    ;;
esac

if [ "$root" != none ]; then
 case "$rootfstype" in
 (btrfs)
   if command -v btrfs >/dev/null; then
     [ -e /dev/btrfs-control ] || modprobe btrfs
     btrfs device scan
   fi
   ;;
 esac
 info "mounting $rootfstype fs on $root (options: $rootflags)"
 chk mount -n -t $rootfstype -o $rootflags $root /root
 if [ $? != 0 ] && ! grep -q "^[^ ]\\+ /root " /proc/mounts; then
   warn "root filesystem ($rootfstype on $root) is NOT mounted!"
 fi
fi

> /config

# prepare bridge iface
# bridge=y should be used together with bootif
if [ -n "$bridge" -a -n "$bootif" ]; then
  d=/root/etc/systemd/network
  mkdir -p $d
  echo -e >$d/br0.netdev \
   "[NetDev]\\nName=br0\\nKind=bridge\\nMACAddress=$bootif"
  if [ -f $d/$iface.network ]; then
    sed "s/^Name=$iface/Name=br0/" $d/$iface.network
  else
    echo -e \
      "[Match]\\nName=br0\\n[Network]\\nDHCP=yes\\n[DHCP]\\nUseDomains=yes"
  fi >$d/br0.network
  echo -e >$d/$iface.network \
    "[Match]\\nName=$iface\\n[Network]\\nBridge=br0"
fi

# gethttp url destination "description"
gethttp() {
  url="$1" dest="$2" descr="$3"
  if [ -z "$bootif" ]; then
    echo -e '\n\nWARNING: no network has been configured, but http requested\n\n'
  else
    n=0
    while ip link ls dev $iface | grep -q NO-CARRIER && [ $n -lt 100 ]; do
      [ $n != 0 ] || echo -e "\\n\\nWaiting for $iface to become ready...\\n"
      sleep .1; n=$(($n+1))
    done
  fi
  echo -e "\\n\\nLoading $descr $url...\n"
  chk wget -O $dest "$url" 2>/dev/console
}

if [ -n "$rootsquash" ]; then
# handle rootsquash=foo.squashfs and ovl=foo.tar

  # setting up overlayfs: various parts:
  # /mnt/boot: boot media
  # /mnt/root: root fs (squashfs)
  # root.lnx: the squashfs image
  # /mnt/ovl: overlay tmpfs
  # /ovl: the new root

  mkdir -p /mnt/root /mnt/ovl

  # first setup overlay tmpfs (with local data)
  chk mount -t tmpfs -o mode=0755 root-ovl-rw /mnt/ovl
  mkdir -p /mnt/ovl/fs /mnt/ovl/wk

  case "$ovl" in
    ( http:* )
      out=/tmp/ovl.tar
      gethttp "$ovl" $out "overlay files"
      ovl=$out
      ;;
    ( * ) ovl=/root/$ovl ;;
  esac
  chk tar -x -C /mnt/ovl/fs -f $ovl -z

  case "$rootsquash" in
    ( http:* )
      out=/mnt/ovl/root.squashfs
      gethttp "$rootsquash" $out "root filesystem"
      rootsquash=$out
      skipsync=y
      ;;
    ( * ) rootsquash=/root/$rootsquash ;;
  esac

  # now mount squashfs on loop7
  modprobe -q loop
  loop=/dev/loop7
  chk losetup -r $loop $rootsquash

  # prepare syncing it if available
  if [ -z "$skipsync" -a -x /sbin/dmsetup ] && modprobe -q dm_mod; then
    sz=$(blockdev --getsz $loop)
    nm=rootsquash
    # dmsetup by default waits for udev so we should have rules..
    dmsetup --noudevrules --noudevsync \
	create $nm --table "0 $sz linear $loop 0"
    rootdev=/dev/mapper/$nm
    [ -e $rootdev ] ||
      ln -s ../dm-0 $rootdev
  else
    rootdev=$loop
  fi

  echo Mounting squashfs...
  chk mount -t squashfs $rootdev /mnt/root
  mkdir -p /ovl
  chk mount -t overlay \
    -o lowerdir=/mnt/root,upperdir=/mnt/ovl/fs,workdir=/mnt/ovl/wk,noatime \
    root-ovl /ovl

  # move it all to proper place
  mkdir -p /ovl/mnt/boot /ovl/mnt/root /ovl/mnt/ovl
  mount --move /root /ovl/mnt/boot 2>/dev/null || :
  chk mount --move /mnt/root /ovl/mnt/root
  chk mount --move /mnt/ovl /ovl/mnt/ovl
  chk mount --move /ovl /root

  if [ $rootdev != $loop ]; then
    n=copy-squashfs-root
    f=/root/etc/$n.sh
    mkdir -p ${f%/*}; cat >$f <<EOF
#! /bin/sh
# created by bootrc
td=/mnt/root.squashfs
nl=/dev/loop8
[ -b $rootdev -a -b $loop ] || exit 0
sz512=\$(blockdev --getsz $loop) # in 512 sectors
[ \$sz512 -gt 0 ] || exit 0
set -- \$(stat -f -c '%S %d' \${td%/*}/); dfKB=\$((\$1*\$2/1024))
# require 64Mb free space
[ \$dfKB -gt \$((sz512*512/1024 + 64*1024)) ] || exit 0
if ionice cp $loop \$td; then
  losetup -r \$nl \$td
  dmsetup load $rootdev --table "0 \$sz512 linear \$nl 0"
  dmsetup suspend $rootdev
  dmsetup resume $rootdev
  losetup -d $loop
else
  rm -f \$td
fi
EOF
    f=/root/etc/systemd/system/$n.service
    mkdir -p ${f%/*}; cat >$f <<EOF
[Service]
Type=oneshot
ExecStartPre=sleep 30
ExecStart=sh /etc/$n.sh
EOF
    f=/root/etc/systemd/system/multi-user.target.wants/$n.service
    mkdir -p ${f%/*}; ln -s ../${f##*/} $f
  fi

fi # rootsquash=/ovl=

find_bootrc() {
  def_bootrc=/etc/bootrc
  case "$bootrc" in
    [Nn0]* ) bootrc=; return 1;;
    [Yy1]* ) bootrc=$def_bootrc;;
    "") if [ ! -f /root$def_bootrc ]; then bootrc=; return 1
        else bootrc=$def_bootrc; fi
        ;;
    *) ;; # use whatever value passed on the command line
  esac
}

if find_bootrc; then
  info "executing $bootrc"
  # copy bootrc into ramfs, to allow it to remount /root if needed.
  chk cp -p /root$bootrc /bootrc
  if [ -f /bootrc ]; then
    . /bootrc /root
    echo ". /bootrc /root" >>/config
  else
    warn "unable to access $bootrc, bootrc will not be executed"
  fi
fi

runshell "before switchroot"

udevadm control --exit

if [ ! "$init" ]; then
  init=/sbin/init
  if [ -x /root$init -o -L /root$init ]; then
    # init can be executable or an absolute symlink, hope for the best
    :
  else
    init=/bin/sh
    warn "no init found, trying $init"
  fi
fi
case "$init" in
  *sh) initargs= ;;
esac

if ! grep -q " /dev " /proc/mounts; then
  movedev=
elif grep -q " /root/dev " /proc/mounts; then
  warn "bootrc mounted its own /dev, umounting initramfs one"
  movedev="umount -n /dev"
else
  movedev="mount -n --move /dev /root/dev || umount -n /dev"
fi

if ! grep -q " /run " /proc/mounts; then
  :
elif grep -q " /root/run " /proc/mounts; then
  umount -n /run
elif [ -d /root/run ]; then
  [ -d /run/tmp ] || mkdir -m 01777 /run/tmp
  [ -d /run/lock ] || mkdir -m 01777 /run/lock
  mount -n --move /run /root/run || umount /run
else
  umount /run
fi

echo "
umount /proc
umount /sys
$movedev
exec /bin/switch_root /root $init $initargs
" >> /config

:
